// Draw textured torus knot + vertex lightening effect
// same as demo 9 but use different vertex/index format and vertex lightening instead of face lightening
// use one object, light position is fixed
// use CImport.InitTL() function to load and init mesh and textures.

#include "defines.h"
#ifdef DEMO10

#include "libutil.h"
#include "quark.h"
#include "armlet.h"
#include "meshdata.h"                            // from AseToMsh directory
#include "import.h"                              // from Demo directory

#define KNOT_SCALE 320                           // radius of torus knot

CVGpu VGpu;                                      // create a virtual gpu object
CTextureList TextureList;                        // use a list of textures
CImport Import;                                  // CImport object used to load mesh + textures from file

int LightPos[3];                                 // position of light
int LightPower;                                  // power of light

int DisplayMode;                                 // used to change display mode (textured/untextured/wireframe)
int MessageTime;                                 // manage time of displayed message
char *Message;                                   // current message

// create a structure to work with 3D objects
typedef struct
{
  CTriangleList Tl;                              // triangle list

  CMatrix Matrix;                                // a matrix to move object
  int ARotations[3];                             // first rotation

  // lightening
  short *VNormalList;                            // contain normal of each vertex, used for lightening
} T3DObject;

T3DObject Obj3D;

// build vertex normal vector using center of object to vertex vector (like on sphere)
void DefCenterVertexNormal(void)
{
  short *NList = Obj3D.VNormalList;
  for (int i=0; i<Obj3D.Tl.VertexListSize; i++)
  {
    int N[3];                                    // vertex position
    Obj3D.Tl.GetXYZ(i, N);                       // get vertex position in triangle list = Normal (torus knot origin = 0,0,0)
 
    Normalize(N);                                // normalize vector (to get norm = TRIGO_MUL)
    NList[0] = N[0];                             // store result in list
    NList[1] = N[1];
    NList[2] = N[2];
    NList += 3;
  }
}

// main game init
int InitGame(void)
{
  if (!VGpu.Init(CLIPZ_NEAR, CLIPZ_FAR, 70))     // init vgpu, Fov = 70 degree (2*PI = 360 degree)
    return 0;

  if (!TextureList.Alloc(1))                     // alloc list for 2 textures
    return 0;

  // --------------------------------------------------------
  // init knot object, Torus knot with 1000 vertex/2000 faces
  // use intensity parameter in vertex format.

  CTriangleList *Tl = &Obj3D.Tl;
  if (!Import.InitTL("Mesh/TKnot2000.msh", Tl, VF_XYZ|VF_I, IF_ABC|IF_UV, KNOT_SCALE, &TextureList, 0))
    return 0;

  // alloc memory to store vertex normal to light knot
  short *NList = (short *)malloc(Tl->VertexListSize*3*sizeof(short));
  if (!NList)
    return 0;
  Obj3D.VNormalList = NList;                     // set ptr into object

  // define initial rendering mode
  Tl->BuildVNormalList(Obj3D.VNormalList, 3);    // use triangle list function

  // messages
  DisplayMode = 1;                               // initial diplayed mode
  MessageTime = 0;                               // manage time of displayed message
  Message = NULL;                                // no message

  LightPos[0] = -100;                            // set light position
  LightPos[1] = 50;
  LightPos[2] = -100;
  LightPower = 255;                              // power of light

  // reserve some working memory for internal work of DrawTriangleList() function.
  if (!VGpu.SetWorkMemory(Tl->VertexListSize, 0))  // max vertex list size is 1000 vertex, 0 sorted faces (alpha) are used
    return 0;

  // reset frame buffer content to init z buffer and avoid garbage displayed on first frame.
  VGpu.ResetFB(COLOR_RGB(0, 0, 64));
  return 1;                                      // init success
}

// scan key to change rendering mode
void ScanKeyboard(void)
{
  int i;
  if (ARMData->Keyboard.KeyPressedId)
  {
    DisplayMode++;
    if (DisplayMode == 8)
      DisplayMode = 0;

    switch (DisplayMode)
    {
      case 0:
        VGpu.DrawCfg.TlUnTextured = 0;
        // reset intensity to max for unlighted mode
        for (i=0; i<Obj3D.Tl.VertexListSize; i++)
          Obj3D.Tl.SetI(i, 255);
        Message = "Textured not lighted";
      break;
      case 1:
        Obj3D.Tl.BuildVNormalList(Obj3D.VNormalList, 3);   // use triangle list function
        Message = "Textured light mode 0";
      break;
      case 2:
        DefCenterVertexNormal();
        Message = "Textured light mode 1";
      break;
      case 3:
        VGpu.DrawCfg.WireFrame = 1;
        Message = "Wireframe Mode";
      break;
      case 4:
        VGpu.DrawCfg.NoCulling = 1;
        Message = "Culling OFF";
      break;
      case 5:
        VGpu.DrawCfg.NoCulling = 0;
        VGpu.DrawCfg.WireFrame = 0;
        VGpu.DrawCfg.TlUnTextured = 1;
        // reset intensity to max for unlighted mode
        for (i=0; i<Obj3D.Tl.VertexListSize; i++)
          Obj3D.Tl.SetI(i, 255);
        Message = "Untextured not lighted";
      break;
      case 6:
        Obj3D.Tl.BuildVNormalList(Obj3D.VNormalList, 3);
        Message = "Untextured light mode 0";
      break;
      case 7:
        DefCenterVertexNormal();
        Message = "Untextured light mode 1";
      break;
    }

    MessageTime = 100;                           // display for 100 frames
    ARMData->Keyboard.KeyPressedId = 0;
  }
  
  if (MessageTime)
  {
    if (Message)
      VGpu.PrintXY(Message, 0, 15);
    MessageTime--;
  }
}

// main game loop, time is passed, must return 0 to quit
int LoopGame(void)
{
  ScanKeyboard();                                // something to change ?

  // -----------------------
  // set torus knot position
  Obj3D.Matrix.LoadIdentity();                   // init matrix
  Obj3D.Matrix.Translate(0, 0, -300);            // define translation on -z axis (camera look on -z direction)
  Obj3D.Matrix.RotateX(Obj3D.ARotations[0]);     // define xyz ARotations
  Obj3D.Matrix.RotateY(Obj3D.ARotations[1]);
  Obj3D.Matrix.RotateZ(Obj3D.ARotations[2]);

  // update Rotations angles
  Obj3D.ARotations[0] += 15;
  Obj3D.ARotations[1] += 20;
  Obj3D.ARotations[2] += 35;

  // --------------------------------------------------------------
  // define lightening intensity on torus knot using light position
  int *B = Obj3D.Matrix.Pos();                   // get position of torus knot

  // light direction
  int AB[3];                                     // light direction
  AB[0] = B[0] - LightPos[0];
  AB[1] = B[1] - LightPos[1];
  AB[2] = B[2] - LightPos[2];
  Normalize(AB);                                 // normalize

  CTriangleList *Tl = &Obj3D.Tl;
  unsigned short *IList = (unsigned short *)&Tl->VertexList[Tl->ViOfs];  // get ptr on first intensity value in triangle list
  short *NList = Obj3D.VNormalList;              // ptr on normal list

  // ----------------------------------------------------------------------------
  // same code than demo 9, except than light power is fixed and we get intensity
  // value into vertex list instead of index list.

  if (DisplayMode && (DisplayMode != 5))         // check if it is a lighted mode
  {
#if 0
    // slowest method: rotate each normal vector of object to light
    for (i=0; i<Tl->VertexListSize; i++)
    {
      int N[3];
      N[0] = NList[0];
      N[1] = NList[1];
      N[2] = NList[2];

      Obj3D.Matrix.RotateVector(N, N);           // perform only rotation using torus knot matrix (ignore translations)
#else
    // fastest method: inverse rotate only light direction, then do not need to rotate each face normal of object to light
    Obj3D.Matrix.InvRotateVector(AB, AB);        // perform inverse rotation using torus knot transposed matrix
    for (int i=0; i<1000; i++)
    {
      int N[3];
      N[0] = NList[0];
      N[1] = NList[1];
      N[2] = NList[2];

#endif
      int Dot = AB[0]*N[0] + AB[1]*N[1] + AB[2]*N[2];  // dot product

      int Intensity = 55;                        // define base 'unlighted' intensity
      if (Dot < 0)                               // triangle is oriented to light
      {
        int di = -Dot >> (TRIGO_SHIFT*2 - 8);    // devide by norm of NList[i] + norm of AB, and scale result by 256, this give ligt effect
        Intensity += (di*200) >> 8;              // 200 = delta light variation du to light position

        // saturate. note: due to lack of accuracy in normalization, di dot product value may be > 256, and produce an intensity value > 255.
        // This will reset low 8 bits, and because 8 hight bits value is ignored, this define black color if saturation check is not done.
        if (Intensity > 255)
          Intensity = 255;
      }

      *IList = Intensity;

      IList += Tl->VertexStride;
      NList += 3;
    }
  }

  // ----
  // draw
  VGpu.ViewMatrix = &Obj3D.Matrix;               // change VGpu view matrix (default is &VGpu.Matrix)
  VGpu.DrawTriangleList(&Obj3D.Tl, &TextureList);  // note: return value of DrawTriangleList ignored here

  // text infos
  if ((VGpu.FrameCounter < 300) && !Message)
  {
    VGpu.PrintXY("Vertex Lightening demo", 0, 15);
    VGpu.PrintXY("Torus Knot: 2000 faces", 0, 30);
    VGpu.PrintXY("Press any key to change mode", 0, 45);
  }
  VGpu.PrintFps();                               // display framerate
  return !ARMData->Mouse.LButtonDown;            // loop util click/tap on screen occur
}

// blit datas into video memory, and clear virtual screen in dark blue
void BlitScreen(void)
{
  VGpu.FlushFB(COLOR_RGB(0, 0, 64));
}

// free memory
void EndGame(void)
{
  Obj3D.Tl.Delete();
  free(Obj3D.VNormalList);
  TextureList.Delete();
  VGpu.Delete();
}
#endif // DEMOx
