// Draw textured spindle with triangle list using 6 different textures

#include "defines.h"
#ifdef DEMO11

#include "libutil.h"
#include "quark.h"
#include "armlet.h"
#include "meshdata.h"
#include "import.h"

#define SPINDLE_SCALE 200                        // scale of spindle
#define TEX_COUNT 6                              // number 1 to 5, + spindle cap

CVGpu VGpu;                                      // create a virtual gpu object
CTextureList TextureList;                        // use a list of textures
CImport Import;                                  // CImport object used to load mesh + textures from file

int LightPos[3];                                 // position of light

// create a structure to work with 3D objects
typedef struct
{
  CTriangleList Tl;                              // triangle list

  CMatrix Matrix;                                // a matrix to move object
  int ARotations[3];                             // rotations
  int Phase;                                     // used to move spindle base and define tilt angles

  // lightening
  short *VNormalList;                            // contain normal of each vertex, used for lightening
} T3DObject;

T3DObject Obj3D;

// main game init
int InitGame(void)
{
  if (!VGpu.Init(CLIPZ_NEAR, CLIPZ_FAR, 70))     // init vgpu, Fov = 70 degree (2*PI = 360 degree)
    return 0;

  if (!TextureList.Alloc(TEX_COUNT))             // alloc list for 2 textures
    return 0;

  // ----------------------------------------------------
  // init spindle object, spindle with 22 vertex/40 faces
  CTriangleList *Tl = &Obj3D.Tl;

  if (!Import.InitTL("Mesh/Spindle.msh", Tl, VF_XYZ|VF_I, IF_ABC|IF_UV, SPINDLE_SCALE, &TextureList, 1))
    return 0;

  // alloc memory to store vertex normal to light spindle
  Obj3D.VNormalList = (short *)malloc(Tl->VertexListSize*3*sizeof(short));
  if (!Obj3D.VNormalList)
    return 0;

  if (!Tl->BuildVNormalList(Obj3D.VNormalList, 3))
    return 0;
 
  // light position
  LightPos[0] = -100;                            // set light position
  LightPos[1] = 150;
  LightPos[2] = -200;

  // reserve some working memory for internal work of DrawTriangleList() function.
  if (!VGpu.SetWorkMemory(Tl->VertexListSize, 0)) // max vertex list size is 22 vertex, 0 sorted faces (alpha) are used
    return 0;

  VGpu.ResetFB(COLOR_RGB(0, 0, 64));
  return 1;                                      // init success
}

// main game loop, time is passed, must return 0 to quit
int LoopGame(void)
{
  // ------------------------------------------------------------------
  // set spindle position, use phase to modify position and tilt angles
  Obj3D.ARotations[1] += 180;                    // self y rotation

  // define x and z tilt angles
  Obj3D.Phase += 70;
  int s = SINUS(Obj3D.Phase);                    // +/- TRIGO_MUL variation
  int c = SINUS(Obj3D.Phase>>2);                 // +/- TRIGO_MUL variation
  int tx = (s*DEGREE(30)) >> TRIGO_SHIFT;        // +/- 30 degree variation (in system angles)
  Obj3D.ARotations[0] = tx;                      // define x tilt

  int tz = (c*DEGREE(10)) >> TRIGO_SHIFT;        // +/- 10 degree variation (in system angles)
  Obj3D.ARotations[2] = tz;                      // define z tilt

  // define base move
  int dx = (s*60) >> TRIGO_SHIFT;                // delta move on x (+/- 60)
  int dz = (c*50) >> TRIGO_SHIFT;                // delta move on z (+/- 50)

  Obj3D.Matrix.LoadIdentity();                   // init matrix
  Obj3D.Matrix.Translate(dx, -120, -320 + dz);   // define translation on -z axis (camera look on -z direction)
  Obj3D.Matrix.RotateX(Obj3D.ARotations[0]);     // x tilt
  Obj3D.Matrix.RotateY(Obj3D.ARotations[1]);     // self rotate on y axis
  Obj3D.Matrix.RotateX(Obj3D.ARotations[2]);     // z tilt

  // --------------------------------------------------------------
  // define lightening intensity on spindle using light position
  int *B = Obj3D.Matrix.Pos();                   // get position of spindle

  // light direction
  int AB[3];                                     // light direction
  AB[0] = B[0] - LightPos[0];
  AB[1] = B[1] - LightPos[1];
  AB[2] = B[2] - LightPos[2];
  Normalize(AB);                                 // normalize

  CTriangleList *Tl = &Obj3D.Tl;
  unsigned short *IList = (unsigned short *)&Tl->VertexList[Tl->ViOfs];  // get ptr on first intensity value in triangle list
  short *NList = Obj3D.VNormalList;              // ptr on normal list

  // ----------
  // lightening
  Obj3D.Matrix.InvRotateVector(AB, AB);          // perform inverse rotation using spindle transposed matrix
  for (int i=0; i<22; i++)
  {
    int N[3];
    N[0] = NList[0];
    N[1] = NList[1];
    N[2] = NList[2];

    int Dot = AB[0]*N[0] + AB[1]*N[1] + AB[2]*N[2];  // dot product
    int Intensity = 100;                         // define base 'unlighted' intensity
    if (Dot < 0)                                 // triangle is oriented to light
    {
      int di = -Dot >> (TRIGO_SHIFT*2 - 8);      // devide by norm of NList[i] + norm of AB, and scale result by 256, this give ligt effect
      Intensity += (di*155) >> 8;                // 155 = delta light variation du to light position

      // saturate
      if (Intensity > 255)
        Intensity = 255;
    }
    *IList = Intensity;
    IList += Tl->VertexStride;
    NList += 3;
  }

  // ----
  // draw
  VGpu.ViewMatrix = &Obj3D.Matrix;               // set VGpu view matrix
  VGpu.DrawTriangleList(&Obj3D.Tl, &TextureList);  // note: return value of DrawTriangleList ignored here

  // text infos
  if (VGpu.FrameCounter < 300)
  {
    VGpu.PrintXY("Single triangle list", 0, 15);
    VGpu.PrintXY("using 6 textures", 0, 30);
  }
  VGpu.PrintFps();                               // display framerate
  return !ARMData->Mouse.LButtonDown;            // loop util click/tap on screen occur
}

// blit datas into video memory, and clear virtual screen in dark blue
void BlitScreen(void)
{
  VGpu.FlushFB(COLOR_RGB(0, 0, 64));
}

// free memory
void EndGame(void)
{
  Obj3D.Tl.Delete();
  free(Obj3D.VNormalList);
  TextureList.Delete();
  VGpu.Delete();
}
#endif // DEMOx
