// Draw textured home with some alpha and alpha keyed faces

#include "defines.h"
#ifdef DEMO12

#include "libutil.h"
#include "quark.h"
#include "armlet.h"
#include "meshdata.h"
#include "import.h"

#define HOME_SCALE 300                           // scale of home
#define TEX_COUNT 5

CVGpu VGpu;                                      // create a virtual gpu object
CTextureList TextureList;                        // use a list of textures
CImport Import;                                  // CImport object used to load mesh + textures from file

const int TexDrawMode[TEX_COUNT] = { RM_OPAQUE, RM_ALPHA, RM_ALPHA, RM_KEYED, RM_OPAQUE };  // define draw mode for each texture
const int Tex2Sided[TEX_COUNT]   = { 0, 1, 1, 1, 0 };  // define witch texture are 2 sided (set for alpha and alpha keyed)

int LightPos[3];                                 // position of light

int DisplayMode;                                 // used to change display mode (textured/untextured/wireframe)
char *Message;                                   // current message
int MessageTime;                                 // manage time of displayed message

// create a structure to work with 3D objects
typedef struct
{
  CTriangleList Tl;                              // triangle list

  CMatrix Matrix;                                // a matrix to move object
  int RotY;                                      // y rotation
  int RotYSpd;                                   // y rotation speed

  // lightening
  short *VNormalList;                            // contain normal of each vertex, used for lightening
} T3DObject;

T3DObject Obj3D;

// main game init
int InitGame(void)
{
  int i;
  if (!VGpu.Init(CLIPZ_NEAR, CLIPZ_FAR, 60))     // init vgpu, Fov = 60 degree (2*PI = 360 degree)
    return 0;

  if (!TextureList.Alloc(TEX_COUNT))             // alloc list for 2 textures
    return 0;

  // ----------------------------------------------------
  // init home object, home with 88 vertex/158 faces
  CTriangleList *Tl = &Obj3D.Tl;

  if (!Import.InitTL("Mesh/Home0.msh", Tl, VF_XYZ|VF_I, IF_ABC|IF_UV, HOME_SCALE, &TextureList, 1))
    return 0;

  // alloc memory to store vertex normal to light home
  Obj3D.VNormalList = (short *)malloc(Tl->VertexListSize*3*sizeof(short));
  if (!Obj3D.VNormalList)
    return 0;

  if (!Tl->BuildVNormalList(Obj3D.VNormalList, 3))
    return 0;
 
  // define rasterization modes for textures
  for (i=0; i<TEX_COUNT; i++)
  {
    CTexture *Tex = &TextureList.Tex[i];

    Tex->RasterMode = TexDrawMode[i];            // set rasterization mode
    Tex->Vis2Sided = Tex2Sided[i];               // alpha faces in this demos must be visible from both side
  }

  // light position
  LightPos[0] = -100;                            // set light position
  LightPos[1] = 100;
  LightPos[2] = -150;

  Obj3D.RotY = 0;                                // initial rotations
  Obj3D.RotYSpd = 20;

  // reserve some working memory for internal work
  //  - 88 vertex for DrawTriangleList() function.
  //  - reserve for 128 sorted faces (6 is enough for this demo)
  if (!VGpu.SetWorkMemory(Tl->VertexListSize, 128))
    return 0;

  VGpu.DrawCfg.ZCorrection = 1;                  // enable perspective correction
  VGpu.DrawCfg.TlSortAlpha = 1;                  // enable z sort of alpha faces
  VGpu.ResetFB(COLOR_RGB(0, 0, 64));             // clear frame buffer for 1st frame
  VGpu.SortReset();                              // empty sort buffer

  DisplayMode = 0;

  return 1;                                      // init success
}

// scan key to change rendering mode
void ScanKeyboard(CVGpu *VGpu)
{
  if (ARMData->Keyboard.KeyPressedId)
  {
    DisplayMode++;
    if (DisplayMode == 4)
      DisplayMode = 0;

    if (DisplayMode == 0)
    {
      VGpu->DrawCfg.TlUnTextured = 1;
      Message = "Untextured Mode";
    }
    else
    if (DisplayMode == 1)
    {
      VGpu->DrawCfg.WireFrame = 1;
      Message = "Wireframe Mode";
    }
    else
    if (DisplayMode == 2)
    {
      VGpu->DrawCfg.NoCulling = 1;
      Message = "Culling OFF";
    }
    else
    if (DisplayMode == 3)
    {
      VGpu->DrawCfg.NoCulling = 0;
      VGpu->DrawCfg.WireFrame = 0;
      VGpu->DrawCfg.TlUnTextured = 0;
      Message = "Textured mode";
    }

    MessageTime = 100;                           // display for 100 frames
    ARMData->Keyboard.KeyPressedId = 0;          // reset
  }
  
  if (MessageTime)
  {
    if (Message)
      VGpu->PrintXY(Message, 0, 30);
    MessageTime--;
  }
}

// main game loop, time is passed, must return 0 to quit
int LoopGame(void)
{
  ScanKeyboard(&VGpu);                           // to change drawing modes

  Obj3D.RotY += Obj3D.RotYSpd;                   // self y rotation
  
  // ------------
  // set position
  Obj3D.Matrix.LoadIdentity();                   // init matrix
  Obj3D.Matrix.Translate(0, -50, -300);          // define translation on -z axis (camera look on -z direction)
  Obj3D.Matrix.RotateY(Obj3D.RotY);              // self rotate on y axis

  // --------------------------------------------------------------
  // define lightening intensity on home using light position
  int *B = Obj3D.Matrix.Pos();                   // get position of home

  // light direction
  int AB[3];                                     // light direction
  AB[0] = B[0] - LightPos[0];
  AB[1] = B[1] - LightPos[1];
  AB[2] = B[2] - LightPos[2];
  Normalize(AB);                                 // normalize

  CTriangleList *Tl = &Obj3D.Tl;
  unsigned short *IList = (unsigned short *)&Tl->VertexList[Tl->ViOfs];  // get ptr on first intensity value in triangle list
  short *NList = Obj3D.VNormalList;              // ptr on normal list

  // ----------
  // lightening
  Obj3D.Matrix.InvRotateVector(AB, AB);          // perform inverse rotation using home transposed matrix
  for (int i=0; i<88; i++)
  {
    int N[3];
    N[0] = NList[0];
    N[1] = NList[1];
    N[2] = NList[2];

    int Dot = AB[0]*N[0] + AB[1]*N[1] + AB[2]*N[2];  // dot product
    int Intensity = 100;                         // define base 'unlighted' intensity
    if (Dot < 0)                                 // triangle is oriented to light
    {
      int di = -Dot >> (TRIGO_SHIFT*2 - 8);      // devide by norm of NList[i] + norm of AB, and scale result by 256, this give ligt effect
      Intensity += (di*155) >> 8;                // 155 = delta light variation du to light position

      // saturate
      if (Intensity > 255)
        Intensity = 255;
    }
    *IList = Intensity;
    IList += Tl->VertexStride;
    NList += 3;
  }

  // ----
  // draw
  VGpu.ViewMatrix = &Obj3D.Matrix;               // set VGpu view matrix
  VGpu.DrawTriangleList(&Obj3D.Tl, &TextureList);  // note: return value of DrawTriangleList ignored here

  VGpu.SortDraw();                               // draw Z sorted faces (alpha faces)

  // text infos
  if ((VGpu.FrameCounter < 150) && !Message)
  {
    VGpu.PrintXY("Alpha transparency + Z Sort", 0, 15);
    VGpu.PrintXY("Press any key to change display mode", 0, 30);
  }

  VGpu.PrintFps();                               // display framerate
  return !ARMData->Mouse.LButtonDown;            // loop util click/tap on screen occur
}

// blit datas into video memory, and clear virtual screen in dark blue
void BlitScreen(void)
{
  VGpu.FlushFB(COLOR_RGB(0, 0, 64));
}

// free memory
void EndGame(void)
{
  Obj3D.Tl.Delete();
  free(Obj3D.VNormalList);
  TextureList.Delete();
  VGpu.Delete();
}
#endif // DEMOx
