// Draw sprite demo

#include "defines.h"
#ifdef DEMO2
#include "libutil.h"
#include "quark.h"
#include "tgafile.h"
#include "armlet.h"                              // required for STR_AP() macro

CVGpu VGpu;                                      // create a virtual gpu object

#define TEX_COUNT 5                              // number of textures used to define sprites
#define SPR_COUNT 16                             // number of sprites to draw

CTexture TextureList[TEX_COUNT];
const char *TexNames[TEX_COUNT] = { "Bird.tga", "Heart.tga", "Frog.tga", "Glass.tga", "Ghostk.tga" };
const int TexDrawMode[TEX_COUNT] = { RM_OPAQUE, RM_KEYED, RM_KEYED, RM_ALPHA, RM_ALPHA | RM_KEYED };

typedef struct
{
  int x, y;                                      // position
  int vx, vy;                                    // speed
  int DrawMode;                                  // drawing mode use to rasterize texture of sprite
  CTexture *Tex;                                 // texture used to draw the sprite
} TSprite;

TSprite SpriteList[SPR_COUNT];

// main game init, must return 1 if success
int InitGame(void)
{
  if (!VGpu.Init(CLIPZ_NEAR, CLIPZ_FAR, 60))     // init vgpu
    return 0;

  // define color key used into textures, rgb = 255, 0, 255 is usually used to define transparent color
  int ColorKey = COLOR_RGB(255, 0, 255);

  // load textures from files
  char Path[80];
  int i, x, y;
  unsigned char *Pixels;

  for (i=0; i<TEX_COUNT; i++)
  {
    strcpy(Path, "Tga24/");                      // path of tga images
    strcat(Path, PICSTR(TexNames[i]));

    if (!LoadTga(Path, &x, &y, &Pixels))         // load a tga file to define texture
      return 0;
    FlipImageY(Pixels, x, y);                    // flip image in y direction (required for sprites)

    CTexture *T = &TextureList[i];
    T->Init(Pixels, x, y, Path);                 // init the texture
    T->DefColorKey(ColorKey);                    // define the color key (do nothing if ColorKey = 0)
    T->RasterMode = TexDrawMode[i];              // define rasterization mode
  }

  // init all sprites
  for (i=0; i<SPR_COUNT; i++)
  {
    TSprite *Spr = &SpriteList[i];

    // define random initial position
    // note: the position is scaled by 256 (left shift by 8)
    Spr->x = (rand() % SCR_X) << 8;
    Spr->y = (rand() % SCR_Y) << 8;

    // define random initial speed, ensure a minimum value
    do
    {
      Spr->vx = -256 + (rand() & 0x1ff);
    } while ((Spr->vx*Spr->vx) < 64*64);

    do
    {
      Spr->vy = -256 + (rand() & 0x1ff);
    } while ((Spr->vy*Spr->vy) < 64*64);

    
    int TexIndex = i % TEX_COUNT;                // define a texture index in list
    Spr->Tex = &TextureList[TexIndex];           // assign a texture
    Spr->DrawMode = TexDrawMode[TexIndex];       // define a drawing mode for texture
  }

  // change default color for text to yellow
  VGpu.DrawCfg.TextColor = COLOR_RGB(255, 255, 0);

  return 1;                                      // init success
}

// main game loop, time is passed, must return 0 to quit
int LoopGame(void)
{
  int i;

  // update position of each sprite
  for (i=0; i<SPR_COUNT; i++)
  {
    TSprite *Spr = &SpriteList[i];

    // get draw position of sprite
    int xL = Spr->x >> 8;                        // x Left
    int xR = xL + Spr->Tex->SizeX;               // x Right
    int yT = Spr->y >> 8;                        // y Top
    int yB = yT + Spr->Tex->SizeY;               // y bottom

    // inverse speed if sprite go out of the screen limits (with 10 pixels margin)
    if ((xL < -10) && (Spr->vx < 0))
      Spr->vx *= -1;

    if ((yT < -10) && (Spr->vy < 0))
      Spr->vy *= -1;

    if ((xR >= (SCR_X + 10)) && (Spr->vx > 0))
      Spr->vx *= -1;

    if ((yB >= (SCR_Y + 10)) && (Spr->vy > 0))
      Spr->vy *= -1;

    // update position using speed
    Spr->x += Spr->vx;
    Spr->y += Spr->vy;

    // draw opaque and alpha keyed sprites only
    if (Spr->Tex->RasterMode < RM_ALPHA)
      VGpu.DrawSprite(Spr->Tex, xL, yT);         // draw

    if (!i)                                      // draw text one time
      VGpu.PrintXY("Sprite Drawing Demo", xL - 80, yT - 16);
  }

  // draw alpha sprites at end to draw them over opaque sprites
  for (i=SPR_COUNT-1; i>=0; i--)
  {
    TSprite *Spr = &SpriteList[i];

    if (Spr->Tex->RasterMode >= RM_ALPHA)
      VGpu.DrawSprite(Spr->Tex, Spr->x >> 8, Spr->y >> 8); // draw
  }

  VGpu.PrintFps();                               // display framerate
  return !ARMData->Mouse.LButtonDown;            // loop util click/tap on screen occur
}

// blit datas into video memory, and clear virtual screen in dark blue
void BlitScreen(void)
{
  VGpu.FlushFB(COLOR_RGB(0, 0, 64));
}

// called before to quit (click on screen occured, or escape pressed, or LoopGame return 0)
void EndGame(void)
{
  for (int i=0; i<TEX_COUNT; i++)                // free texture memory
    free(TextureList[i].Texels);

  VGpu.Delete();                                 // free VGpu memory
}
#endif // DEMOx
