// Scaled sprite demo

#include "defines.h"
#ifdef DEMO3
#include "libutil.h"
#include "quark.h"
#include "tgafile.h"
#include "armlet.h"                              // required for PICSTR() macro

CVGpu VGpu;                                      // create a virtual gpu object

#define TEX_COUNT 6                              // 5 peoples and one tree
#define SPR_COUNT 20                             // numbrer of sprites to draw

CTexture TextureList[TEX_COUNT];
const char *TexNames[TEX_COUNT] = { "People0.tga", "People1.tga", "People2.tga", "People3.tga", "People4.tga", "Tree0.tga" };
int TexScale[TEX_COUNT] = { 256, 256, 256, 256, 256, 600 };

// we simulate a 3D position of sprite to zoom sprite depending of z distance
typedef struct
{
  int x, y, z;                                   // x, y, z, position
  int Scale;                                     // default scale of sprite
  CTexture *Tex;                                 // texture used to draw the sprite
} TSprite;

TSprite SpriteList[SPR_COUNT];

// struct used to sort sprites depending of distance
typedef struct
{
  int Dist;                                      // distance
  int Index;                                     // index of sprite in SpriteList
} TSprSort;

TSprSort SprSort[SPR_COUNT];

// main game init, must return 1 if success
int InitGame(void)
{
  if (!VGpu.Init(CLIPZ_NEAR, CLIPZ_FAR, 60))     // init vgpu
    return 0;

  // define color key, black used in this demo
  int ColorKey = COLOR_RGB(0, 0, 0);

  // load textures from files
  char Path[80];
  int i, x, y;
  unsigned char *Pixels;

  for (i=0; i<TEX_COUNT; i++)
  {
    strcpy(Path, "Tga24/");                      // tga images are stored in Data folder
    strcat(Path, PICSTR(TexNames[i]));

    if (!LoadTga(Path, &x, &y, &Pixels))         // load a tga file to define texture
      return 0;
    FlipImageY(Pixels, x, y);                    // flip image in y direction (required with sprites)

    CTexture *T = &TextureList[i];
    T->Init(Pixels, x, y, Path);                 // init the texture
    T->DefColorKey(ColorKey);
    T->RasterMode = RM_KEYED;                    // all sprites in this demo are color keyed
  }

  // init all sprites positions
  int Rw = 30;                                   // width of road
  for (i=0; i<SPR_COUNT; i++)
  {
    TSprite *Spr = &SpriteList[i];

    if (i & 1)
      Spr->x = -Rw;                              // left road border
    else
      Spr->x = Rw;                               // right road border
   
    Spr->y = -30;
    Spr->z = 50 + i*100 + (rand() % 50);

    int TexIndex = rand() % TEX_COUNT;
    Spr->Tex = &TextureList[TexIndex];           // assign a texture
    Spr->Scale = TexScale[TexIndex];             // define a texture depending of scale (want tree size > people size)
  }
  return 1;                                      // init success
}

// distance sort call back for sprites
static int ZSortCB(const void *arg1, const void *arg2)
{
  TSprSort *A = (TSprSort *)arg1;
  TSprSort *B = (TSprSort *)arg2;

  if (A->Dist < B->Dist)
    return 1;

  if (A->Dist > B->Dist)
    return -1;

  return 0;
}

// main game loop, time is passed, must return 0 to quit
int LoopGame(void)
{
  int i;

  // update position of each sprite
  for (i=0; i<SPR_COUNT; i++)
  {
    TSprite *Spr = &SpriteList[i];
    
    Spr->z -= 3;
    if (Spr->z < 20)
      Spr->z = 1000 + (rand() % 80);

    SprSort[i].Dist = Spr->z;
    SprSort[i].Index = i;
  }

  // sort sprites depending of z distance
  // note: sort is used for tutorial reasons, and can be avoid using index of most far sprite in previous loop
  qsort(SprSort, SPR_COUNT, sizeof(TSprSort), ZSortCB);

  // draw
  for (i=0; i<SPR_COUNT; i++)
  {
    int ZIndex = SprSort[i].Index;
    TSprite *Spr = &SpriteList[ZIndex];

    // define 3D to 2D screen position and Scale value depending of sprite distance
    int Scale = (Spr->Scale*150)/Spr->z;
    int xe = (SCR_X/2) + (150*Spr->x)/Spr->z;
    int ye = (SCR_Y/2) - (150*Spr->y)/Spr->z;

    // center sprite on position
    xe -= (Spr->Tex->SizeX * Scale) >> (8 + 1);
    ye -= (Spr->Tex->SizeY * Scale) >> 8;

    // draw sprite
    VGpu.DrawSpriteScale(Spr->Tex, xe, ye, Scale); // draw
  }

  VGpu.PrintXY("Scaled Sprite Demo", 0, 16);
  VGpu.PrintFps();                               // display framerate
  return !ARMData->Mouse.LButtonDown;            // loop util click/tap on screen occur
}

// blit datas into video memory, and clear virtual screen in dark blue
void BlitScreen(void)
{
  VGpu.FlushFB(COLOR_RGB(0, 0, 64));
}

// called before to quit (click on screen occured, or escape pressed, or LoopGame return 0)
void EndGame(void)
{
  for (int i=0; i<TEX_COUNT; i++)                // free texture memory
    free(TextureList[i].Texels);

  VGpu.Delete();                                 // free VGpu memory
}
#endif // DEMOx