// Scaled sprite demo using Z buffer. 3D method to use sprites
// Scaled sprite using Z buffer are slower to draw than normal sprites but allow:
//  - to mix sprites with 3D triangle geometry (billboard)
//  - to avoid a Z sorting of sprites (except if alpha transparency is used)

#include "defines.h"
#ifdef DEMO4
#include "libutil.h"
#include "quark.h"
#include "tgafile.h"
#include "armlet.h"                              // required for STR_AP() macro

CVGpu VGpu;                                      // create a virtual gpu object

#define TEX_COUNT 6                              // 5 peoples and one tree
#define SPR_COUNT 20                             // numbrer of sprites to draw

CTexture TextureList[TEX_COUNT];
const char *TexNames[TEX_COUNT] = { "People0.tga", "People1.tga", "People2.tga", "People3.tga", "People4.tga", "Tree0.tga" };
int SprSize[TEX_COUNT][2] = { {50, 110}, {50, 110}, {50, 110}, {50, 100}, {50, 100}, {200, 400} };  // 3D xy size of sprite

// we simulate a 3D position of sprite to zoom sprite depending of z distance
// scaled sprites require 2 3D coordinate of top left vertex and bottom right vertex
typedef struct
{
  int PosTl[3];                                  // to left 3D position
  int PosBr[3];                                  // bottom right position
  CTexture *Tex;                                 // texture used to draw the sprite
} TSprite;

TSprite SpriteList[SPR_COUNT];

// main game init, must return 1 if success
int InitGame(void)
{
  if (!VGpu.Init(CLIPZ_NEAR, CLIPZ_FAR, 60))     // init vgpu
    return 0;

  // define color key, black used in this demo
  int ColorKey = COLOR_RGB(0, 0, 0);

  // load textures from files
  char Path[80];
  int i, x, y;
  unsigned char *Pixels;

  for (i=0; i<TEX_COUNT; i++)
  {
    strcpy(Path, "Tga24/");                      // tga images are stored in Data folder
    strcat(Path, PICSTR(TexNames[i]));

    if (!LoadTga(Path, &x, &y, &Pixels))         // load a tga file to define texture
      return 0;
    FlipImageY(Pixels, x, y);                    // flip image in y direction (required with sprites)

    CTexture *T = &TextureList[i];
    T->Init(Pixels, x, y, Path);                 // init the texture
    T->DefColorKey(ColorKey);
    T->RasterMode = RM_KEYED;                    // all sprites in this demo are color keyed
  }

  // init all sprites positions (3D world positions)
  int Rw = 100;                                  // width of road
  for (i=0; i<SPR_COUNT; i++)
  {
    TSprite *Spr = &SpriteList[i];
    int TexIndex = i % TEX_COUNT;

    int xOfs, yOfs;
    if (i & 1)
      xOfs = -Rw;                                // left road border
    else
      xOfs = Rw;                                 // right road border
   
    yOfs = -80;

    // z position
    int z = CLIPZ_NEAR - i*300 - (rand() % 50); // use negative world 3D positions, camera look on -Z

    // Top Left position
    Spr->PosTl[0] = xOfs - (SprSize[TexIndex][0]/2); // center on x
    Spr->PosTl[1] = yOfs + SprSize[TexIndex][1];
    Spr->PosTl[2] = z;

    // Bottom right position
    Spr->PosBr[0] = xOfs + (SprSize[TexIndex][0]/2);
    Spr->PosBr[1] = yOfs;                        // align on bottom
    Spr->PosBr[2] = z;
    
    Spr->Tex = &TextureList[TexIndex];           // assign a texture
  }
  return 1;                                      // init success
}

// main game loop, time is passed, must return 0 to quit
int LoopGame(void)
{
  int i;

  // update position of each sprite
  for (i=0; i<SPR_COUNT; i++)
  {
    TSprite *Spr = &SpriteList[i];
    
    // note: can ignore one of the 2 z position, (must be same), only z of PosTl used here
    Spr->PosTl[2] += 15;                         // reduce distance on -z axis
    
    // loop when sprite too near
    if (Spr->PosTl[2] > CLIPZ_NEAR)
      Spr->PosTl[2] = -2500 - (rand() % 100);
  }

  // draw
  for (i=0; i<SPR_COUNT; i++)
  {
    TSprite *Spr = &SpriteList[i];

    // draw sprites only if visibles in frustrum. (show general method, although there are always visibles in this demo)

    int m1, m2;                                  // clipping mask
    // values of clipping mask (returned by ScrProjection() and ClipMsk() functions)
    // m = 0  : into frustrum
    // m & 1  : bottom clipped
    // m & 2  : top clipped
    // m & 4  : left clipped
    // m & 8  : right clipped
    // m & 16 : too far < CLIPZ_FAR
    // m & 32 : too near > CLIPZ_NEAR  -> coordinate are not transformed

    // for example if both mask m1 and m2 have b3 = 1 (value = 4) the 2 screen point
    // of sprite are at left of view frustrum, then sprite is not visible.
    // note: if b5 = 1 (value 32), the 3D position is behind near clip and may be not
    // transformable into 2D position (division/0, or negative distance), function
    // ScrProjection do not return 2D position in this case. (value not modifyed)

    int ScrTl[3];                                // top left screen position
    int ScrBr[3];                                // bottom right screen position

    m1 = VGpu.ScrProjection(Spr->PosTl, ScrTl);  // 3D -> 2D projection
    if (m1 & (32 | 16))
      continue;                                  // too far or too near

    Spr->PosBr[2] = Spr->PosTl[2];               // update z to transform
    m2 = VGpu.ScrProjection(Spr->PosBr, ScrBr);
    if (m1 & m2)
      continue;                                  // sprite is out of screen and not visible

    // draw sprite
    VGpu.DrawSpriteScaleZb(Spr->Tex, ScrTl, ScrBr, ScrTl[2]);  // draw using Z buffer. note ScrTl[2] & ScrBr[2] (z) ignored in function
  }

  VGpu.PrintXY("3D Scaled Sprite (ZBuffer)", 0, 16);
  VGpu.PrintFps();                               // display framerate
  return !ARMData->Mouse.LButtonDown;            // loop util click/tap on screen occur
}

// blit datas into video memory, and clear virtual screen in dark blue
void BlitScreen(void)
{
  VGpu.FlushFB(COLOR_RGB(0, 0, 64));
}

// called before to quit (click on screen occured, or escape pressed, or LoopGame return 0)
void EndGame(void)
{
  for (int i=0; i<TEX_COUNT; i++)                // free texture memory
    free(TextureList[i].Texels);

  VGpu.Delete();                                 // free VGpu memory
}
#endif // DEMOx
