// Draw an untextured cube

#include "defines.h"
#ifdef DEMO6
#include "libutil.h"
#include "quark.h"
#include "armlet.h"

CVGpu VGpu;                                      // create a virtual gpu object

#define CUBE_SIZE 100                            // size of cube

// define vertex cube as follow
//           4 ---- 5
//          /|     /|
// +y      / |    / |
// | -z   6 ---- 7  |
// |/     |  |   |  |
// O--+x  |  0 --|- 1
//        | /    | /
//        |/     |/
//        2 ---- 3

// define vertex positions centered on origin (x=0, y=0, z=0) to draw a cube
const signed char CubeVxDef[8][3] = {
  {-1, -1, -1}, { 1, -1, -1}, {-1, -1,  1}, { 1, -1,  1},
  {-1,  1, -1}, { 1,  1, -1}, {-1,  1,  1}, { 1,  1,  1} };

// define face vertex index, 2 triangles are required for each face
// note: take care or index rotation for culling test
const char FaceIndex[12][3] = {
  {0, 1, 3}, {0, 3, 2},     // bottom faces
  {0, 4, 5}, {0, 5, 1},     // rear faces
  {0, 2, 6}, {0, 6, 4},     // left faces
  {7, 5, 4}, {7, 4, 6},     // up faces
  {7, 3, 1}, {7, 1, 5},     // right faces
  {7, 6, 2}, {7, 2, 3} };   // front faces

int CubeRotations[3];
unsigned char FaceColor[6];                      // color of each face

CMatrix TMatrix;                                 // need to define a matrix to rotate cube

// main game init
int InitGame(void)
{
  if (!VGpu.Init(CLIPZ_NEAR, CLIPZ_FAR, 70))     // init vgpu, Fov = 70 degree (2*PI = 360 degree)
    return 0;

  // init cube face colors
  // note: COLOR_RGB must be used to define base color (do not use COLOR_RGBI, intensity is defined by vertex I parameter)
  for (int i=0; i<6; i++)
    FaceColor[i] = COLOR_RGB((rand() & 0xff), (rand() & 0xff), (rand() & 0xff));

  // init xyz rotations
  CubeRotations[0] = 0;
  CubeRotations[1] = 0;
  CubeRotations[2] = 0;

  // assign matrix for VGpu
  VGpu.ViewMatrix = &TMatrix;
  return 1;                                      // init success
}

// main game loop, time is passed, must return 0 to quit
int LoopGame(void)
{
  int i;

  // define TVertex list required to draw some triangles
  TVertex CubeVertex[8];

  // define view matrix (VGpu.ViewMatrix == &TMatrix as defined into init)
  VGpu.ViewMatrix->LoadIdentity();               // init VGpu matrix
  VGpu.ViewMatrix->Translate(0, 0, -200);        // define translation on -z axis (camera look on -z direction)
  VGpu.ViewMatrix->RotateX(CubeRotations[0]);    // define xyz rotations
  VGpu.ViewMatrix->RotateY(CubeRotations[1]);
  VGpu.ViewMatrix->RotateZ(CubeRotations[2]);

  // update rotations angles
  CubeRotations[0] += 15;
  CubeRotations[1] += 20;
  CubeRotations[2] += 25;

  // define position of vertex on screen
  int WorldPos[3];                               // position of cube into world
  int ViewPos[3];                                // position of cube when wiew matrix transformation is applyed
  int ScreenPos[3];                              // position of vertex on screen
  for (i=0; i<8; i++)
  {
    // define world position of cube vertex
    // note: could be done one time in init, but must be stored into array
    WorldPos[0] = CubeVxDef[i][0]*(CUBE_SIZE/2);
    WorldPos[1] = CubeVxDef[i][1]*(CUBE_SIZE/2);
    WorldPos[2] = CubeVxDef[i][2]*(CUBE_SIZE/2);

    // define view position using defined ViewMatrix
    VGpu.ViewMatrix->TransformVector(WorldPos, ViewPos);

    // define screen position
    CubeVertex[i].m = VGpu.ScrProjection(ViewPos, ScreenPos);

    // finish vertex init: 
    // note: vertex uv value not used and do not require init in this demo
    CubeVertex[i].x = ScreenPos[0];
    CubeVertex[i].y = ScreenPos[1];
    CubeVertex[i].z = ScreenPos[2];
    CubeVertex[i].i = 255;                       // set vertex color light intensity
  }

  // draw cube triangles
  for (i=0; i<12; i++)
    VGpu.DrawTriangleU(&CubeVertex[(int)FaceIndex[i][0]], 
                       &CubeVertex[(int)FaceIndex[i][1]],
                       &CubeVertex[(int)FaceIndex[i][2]], FaceColor[i >> 1], 0);

  VGpu.PrintXY("Untextured triangle demo", 0, 16);
  VGpu.PrintFps();                               // display framerate
  return !ARMData->Mouse.LButtonDown;            // loop util click/tap on screen occur
}

// blit datas into video memory, and clear virtual screen in dark blue
void BlitScreen(void)
{
  VGpu.FlushFB(COLOR_RGB(0, 0, 64));
}

// called before to quit (click on screen occured, or escape pressed, or LoopGame return 0)
void EndGame(void)
{
  VGpu.Delete();                                 // free VGpu memory
}
#endif // DEMOx
