// Draw textured cube

#include "defines.h"
#ifdef DEMO7
#include "libutil.h"
#include "quark.h"
#include "tgafile.h"
#include "armlet.h"                              // required for PICSTR() macro

CVGpu VGpu;                                      // create a virtual gpu object

#define CUBE_SIZE 100                            // size of cube

// define vertex positions centered on origin (x=0, y=0, z=0) to draw a cube.
// note: the vertex positions are not same as defined in demo 6, because data have been
// generated by a 3D modeler program, this to avoid to define 'by hand' terture coordinates
const signed char CubeVxDef[8][3] = {
  {-1,  1, -1}, { 1,  1, -1}, {-1,  1,  1}, { 1,  1,  1},
  {-1, -1,  1}, { 1, -1,  1}, {-1, -1, -1}, { 1, -1, -1}};

// define face vertex index, 2 triangles are required for each face
// note: take care or index rotation for culling test
const char FaceIndex[12][3] = {
  {1, 0, 2}, {2, 3, 1},
  {4, 5, 3}, {3, 2, 4},
  {0, 6, 4}, {4, 2, 0},
  {6, 7, 5}, {5, 4, 6},
  {6, 0, 1}, {1, 7, 6},
  {7, 1, 3}, {3, 5, 7}};

// texture coordinates for each 3 vertex of each face
const short FaceUvDef[12][3][2] = {
  {{1, 0}, {1, 1}, {0, 1}},
  {{0, 1}, {0, 0}, {1, 0}},
  {{0, 1}, {0, 0}, {1, 0}},
  {{1, 0}, {1, 1}, {0, 1}},
  {{0, 1}, {0, 0}, {1, 0}},
  {{1, 0}, {1, 1}, {0, 1}},
  {{0, 1}, {0, 0}, {1, 0}},
  {{1, 0}, {1, 1}, {0, 1}},
  {{1, 1}, {0, 1}, {0, 0}},
  {{0, 0}, {1, 0}, {1, 1}},
  {{1, 0}, {1, 1}, {0, 1}},
  {{0, 1}, {0, 0}, {1, 0}}};

// define textures to load
// note: texture x and y size must be a power of 2 between 16 to 128 (ex: 32*32, 32*64, etc..)
#define TEX_COUNT 6                              // use a diferent texture for each face

CTexture TextureList[TEX_COUNT];
const char *TexNames[TEX_COUNT] = { "Wood.tga", "Gnd1.tga", "Gnd2.tga", "Glass1.tga", "Lin.tga", "Water.tga" };

int CubeRotations[3];

CMatrix TMatrix;                                 // need to define a matrix to rotate cube

// main game init
int InitGame(void)
{
  if (!VGpu.Init(CLIPZ_NEAR, CLIPZ_FAR, 70))     // init vgpu, Fov = 70 degree (2*PI = 360 degree)
    return 0;

  // load textures from files
  char Path[80];
  int i, x, y;
  unsigned char *Pixels;

  for (i=0; i<TEX_COUNT; i++)
  {
    strcpy(Path, "Tga24/");                      // tga images are stored in Tga24 Data folder
    strcat(Path, PICSTR(TexNames[i]));

    if (!LoadTga(Path, &x, &y, &Pixels))         // load a tga file to define texture
      return 0;

    // init the texture.
    // note: 
    //  - Init return 1 if size x and y are a power of 2 and in 16 to 128 range. This is
    //    required only to draw triangles. Can be ignored to draw sprites (no size constraints in this case)
    if (!TextureList[i].Init(Pixels, x, y, Path))
      return 0;
  }

  // init xyz rotations
  CubeRotations[0] = 0;
  CubeRotations[1] = 0;
  CubeRotations[2] = 0;

  // enable perspective correction
  VGpu.DrawCfg.ZCorrection = 1;

  // assign matrix for VGpu
  VGpu.ViewMatrix = &TMatrix;
  return 1;                                      // init success
}

// main game loop, time is passed, must return 0 to quit
int LoopGame(void)
{
  int i;

  // define TVertex list required to draw some triangles
  TVertex CubeVertex[8];

  // define view matrix (VGpu.ViewMatrix == &TMatrix as defined into init)
  VGpu.ViewMatrix->LoadIdentity();                // init VGpu matrix
  VGpu.ViewMatrix->Translate(0, 0, -200);         // define translation on -z axis (camera look on -z direction)
  VGpu.ViewMatrix->RotateX(CubeRotations[0]);     // define xyz rotations
  VGpu.ViewMatrix->RotateY(CubeRotations[1]);
  VGpu.ViewMatrix->RotateZ(CubeRotations[2]);

  // update rotations angles
  CubeRotations[0] += 15;
  CubeRotations[1] += 20;
  CubeRotations[2] += 25;

  // define position of cube vertex on screen
  int WorldPos[3];                               // position of vertex into world
  int ViewPos[3];                                // position of vertex when wiew matrix transformation is applyed
  for (i=0; i<8; i++)
  {
    // define world position of cube vertex
    // note: could be done one time in init, but must be stored into array
    WorldPos[0] = CubeVxDef[i][0]*(CUBE_SIZE/2);
    WorldPos[1] = CubeVxDef[i][1]*(CUBE_SIZE/2);
    WorldPos[2] = CubeVxDef[i][2]*(CUBE_SIZE/2);

    // define view position using defined ViewMatrix
    VGpu.ViewMatrix->TransformVector(WorldPos, ViewPos);

    // define screen position, and store result in vertex directly (more fast than usind 3rd vector as in demo 6)
    CubeVertex[i].m = VGpu.ScrProjection(ViewPos, &CubeVertex[i].x);  // note &CubeVertex[i].x = int[3]

    // finish vertex init: 
    CubeVertex[i].i = 255;                       // set vertex color light intensity
  }

  // define uv and draw cube triangles
  for (i=0; i<12; i++)
  {
    TVertex *A, *B, *C;
    A = &CubeVertex[(int)FaceIndex[i][0]];       // get vertex used for each triange
    B = &CubeVertex[(int)FaceIndex[i][1]];
    C = &CubeVertex[(int)FaceIndex[i][2]];

    // define uv for each vertex
    // note: use UV_MUL to convert uv from float to short (u = UV_MUL mean u = 1.0f in float)
    A->u = FaceUvDef[i][0][0]*UV_MUL;            // set vertex uv
    A->v = FaceUvDef[i][0][1]*UV_MUL;
    B->u = FaceUvDef[i][1][0]*UV_MUL;
    B->v = FaceUvDef[i][1][1]*UV_MUL;
    C->u = FaceUvDef[i][2][0]*UV_MUL;
    C->v = FaceUvDef[i][2][1]*UV_MUL;

    VGpu.DrawTriangleT(A, B, C, &TextureList[i >> 1]);
  }

  VGpu.PrintXY("Textured DrawTriangle Demo (1)", 0, 16);
  VGpu.PrintFps();                               // display framerate
  return !ARMData->Mouse.LButtonDown;            // loop util click/tap on screen occur
}

// blit datas into video memory, and clear virtual screen in dark blue
void BlitScreen(void)
{
  VGpu.FlushFB(COLOR_RGB(0, 0, 64));
}

// called before to quit (click on screen occured, or escape pressed, or LoopGame return 0)
void EndGame(void)
{
  for (int i=0; i<TEX_COUNT; i++)                // free texture memory
    free(TextureList[i].Texels);

  VGpu.Delete();                                 // free VGpu memory
}
#endif // DEMOx
