// Draw textured cube using triangle list
// Avantages:
//  - more fast than drawing with separate triangles.
//  - faces are cutted with near clip plane if required (ClipNear position).
//  - work directly with world coordinates, no use to tramform 3D -> 2D and manage clipping mask

#include "defines.h"
#ifdef DEMO8
#include "libutil.h"
#include "quark.h"
#include "tgafile.h"
#include "armlet.h"

CVGpu VGpu;                                      // create a virtual gpu object

#define CUBE_SIZE 100                            // size of cube

// define vertex positions centered on origin (x=0, y=0, z=0) to draw a cube.
// note: the vertex positions are not same as defined in demo 6, because data have been
// generated by a 3D modeler program, this to avoid to define 'by hand' texture coordinates
const signed char CubeVxDef[8][3] = {
  {-1,  1, -1}, { 1,  1, -1}, {-1,  1,  1}, { 1,  1,  1},
  {-1, -1,  1}, { 1, -1,  1}, {-1, -1, -1}, { 1, -1, -1}};

// define face vertex index, 2 triangles are required for each face
// note: take care or index rotation for culling test
const unsigned short FaceIndex[12][3] = {
  {1, 0, 2}, {2, 3, 1},
  {4, 5, 3}, {3, 2, 4},
  {0, 6, 4}, {4, 2, 0},
  {6, 7, 5}, {5, 4, 6},
  {6, 0, 1}, {1, 7, 6},
  {7, 1, 3}, {3, 5, 7}};

// texture coordinates for each 3 vertex of each face
const short FaceUvDef[12][3][2] = {
  {{1, 0}, {1, 1}, {0, 1}},
  {{0, 1}, {0, 0}, {1, 0}},
  {{0, 1}, {0, 0}, {1, 0}},
  {{1, 0}, {1, 1}, {0, 1}},
  {{0, 1}, {0, 0}, {1, 0}},
  {{1, 0}, {1, 1}, {0, 1}},
  {{0, 1}, {0, 0}, {1, 0}},
  {{1, 0}, {1, 1}, {0, 1}},
  {{1, 1}, {0, 1}, {0, 0}},
  {{0, 0}, {1, 0}, {1, 1}},
  {{1, 0}, {1, 1}, {0, 1}},
  {{0, 1}, {0, 0}, {1, 0}}};

int CubeRotations[3];

CMatrix TMatrix;                                 // need to define a matrix to rotate cube
CTextureList TextureList;                        // define a texture list. It is required to work with triangle list even if only one textures is used.
CTriangleList CubeTL;                            // triangle list

// main game init
int InitGame(void)
{
  if (!VGpu.Init(CLIPZ_NEAR, CLIPZ_FAR, 70))     // init vgpu, Fov = 70 degree (2*PI = 360 degree)
    return 0;

  if (!TextureList.Alloc(1))                     // alloc texture list to contain one texture
    return 0;

  // load a texture from file
  char Path[80];
  int x, y;
  unsigned char *Pixels;

  strcpy(Path, "Tga24/");
  strcat(Path, "Wood.tga");

  if (!LoadTga(Path, &x, &y, &Pixels))           // load a tga file to define texture
    return 0;

  // init the texture.
  if (TextureList.Tex[0].Init(Pixels, x, y, Path))
    TextureList.ListSize++;
  else
    return 0;

  // create triangle list, alloc memory for vertex and faces depending of defined format
  int VertexFormat = VF_XYZ;                     // store vertex position
  int IndexFormat = IF_ABC | IF_UV;              // store vertex index + vertex uv

  if (!CubeTL.Alloc(VertexFormat, 8, IndexFormat, 12))  // 8 vertex and 12 faces used, one single draw state is used
    return 0;

  // init the vertex list
  int i;
  for (i=0; i<8; i++)
  {
    short Pos[3];
    Pos[0] = CubeVxDef[i][0]*(CUBE_SIZE/2);      // x
    Pos[1] = CubeVxDef[i][1]*(CUBE_SIZE/2);      // y
    Pos[2] = CubeVxDef[i][2]*(CUBE_SIZE/2);      // z
    CubeTL.SetXYZ(i, Pos);
  }

  // init the face list
  for (i=0; i<12; i++)
  {
    // define vertex index
    CubeTL.SetABC(i, FaceIndex[i]);

    // define texture uv
    short uv[6];
    uv[0] = FaceUvDef[i][0][0]*UV_MUL;           // uA
    uv[1] = FaceUvDef[i][0][1]*UV_MUL;           // vA
    uv[2] = FaceUvDef[i][1][0]*UV_MUL;           // uB
    uv[3] = FaceUvDef[i][1][1]*UV_MUL;           // vB
    uv[4] = FaceUvDef[i][2][0]*UV_MUL;           // uC
    uv[5] = FaceUvDef[i][2][1]*UV_MUL;           // vC
    CubeTL.SetUV(i, uv);
  }

  // build command list
  if (!CubeTL.BuildCmdList(NULL, 0))
    return 0;

  // the DrawTriangleList function require some allocated temporary memory to work
  // to alloc this memory, use VGpu.SetWorkMemory function and pass in MaxVertex parameter
  // the maximum number of vectex that will be used in biggest triangle list.
  // here only one triangle list is used with 8 vertex.
  // the second parameter is used to reserve some memory to sort alpha faces using hash
  // table and is not used in this demo.
  // to free allocated memory, call SetWorkMemory(0, 0), or done by VGpu.Delete function.
  if (!VGpu.SetWorkMemory(8, 0))                 // reserve some working memory for internal work
    return 0;

  // init xyz rotations
  CubeRotations[0] = 0;
  CubeRotations[1] = 0;
  CubeRotations[2] = 0;

  VGpu.DrawCfg.ZCorrection = 1;                  // enable perspective correction

  VGpu.ViewMatrix = &TMatrix;                    // assign matrix for VGpu
  return 1;                                      // init success
}

// main game loop, time is passed, must return 0 to quit
int LoopGame(void)
{
  // define view matrix
  VGpu.ViewMatrix->LoadIdentity();               // init VGpu matrix
  VGpu.ViewMatrix->Translate(0, 0, -200);        // define translation on -z axis (camera look on -z direction)
  VGpu.ViewMatrix->RotateX(CubeRotations[0]);    // define xyz rotations
  VGpu.ViewMatrix->RotateY(CubeRotations[1]);
  VGpu.ViewMatrix->RotateZ(CubeRotations[2]);

  // draw
  int ErrorCode = VGpu.DrawTriangleList(&CubeTL, &TextureList);
  if (ErrorCode)                                 // eErrorCode value returned, can be used to debug
    return 0;                                    // terminate program

  // update rotations angles
  CubeRotations[0] += 15;
  CubeRotations[1] += 20;
  CubeRotations[2] += 25;

  VGpu.PrintXY("DrawTriangleList demo", 0, 16);
  VGpu.PrintFps();                               // display framerate
  return !ARMData->Mouse.LButtonDown;            // loop util click/tap on screen occur
}

// blit datas into video memory, and clear virtual screen in dark blue
void BlitScreen(void)
{
  VGpu.FlushFB(COLOR_RGB(0, 0, 64));
}

// free memory
void EndGame(void)
{
  TextureList.Delete();
  CubeTL.Delete();
  VGpu.Delete();
}
#endif // DEMOx
