// Same demo than demo9.cpp but use CImport.InitTL() function.

#include "defines.h"
#ifdef DEMO9b

#include "libutil.h"
#include "quark.h"
#include "tgafile.h"
#include "meshdata.h"
#include "import.h"
#include "armlet.h"

#define SPH_R0 200                               // radius of sphere 0
#define SPH_R1 60                                // radius of sphere 1

CVGpu VGpu;                                      // create a virtual gpu object
CTextureList TextureList;                        // use a list of textures
CImport Import;                                  // CImport object used to load mesh + textures from file

int DisplayMode;                                 // used to change display mode (textured/untextured/wireframe)
char *Message;                                   // current message
int MessageTime;                                 // manage time of displayed message

// create a structure to work with 3D objects
typedef struct
{
  CTriangleList Tl;                              // triangle list used to draw object

  CMatrix Matrix;                                // a matrix to move object
  int ARotations[3];                             // first rotation
  int BRotations[3];                             // second rotation (for satellite sphere)

  // face lightening
  short *FNormalList;                            // contain normal of each face, used for lightening
} T3DObject;

T3DObject Obj3DList[2];                          // use 2 spheres, one with 180 face, other one with 80 face

// main game init
int InitGame(void)
{
  int i;
  if (!VGpu.Init(CLIPZ_NEAR, CLIPZ_FAR, 70))     // init vgpu, Fov = 70 degree (2*PI = 360 degree)
    return 0;

  if (!TextureList.Alloc(2))                     // alloc list for 2 textures
    return 0;

  // -------------------------------------------------
  // load object 0, geosphere with 92 vertex/180 faces

  // load a sphere mesh from file + texture + create triangle list
  CTriangleList *Tl = &Obj3DList[0].Tl;
  if (!Import.InitTL("Mesh/Sph180.msh", Tl, VF_XYZ, IF_ABC | IF_I | IF_UV, SPH_R0, &TextureList, 0))
    return 0;

  // alloc memory to store face normal to light geosphere 0
  short *NList = (short *)malloc(Tl->IndexListSize*3*sizeof(short));
  if (!NList)
    return 0;
  Obj3DList[0].FNormalList = NList;              // set ptr into object

  // init face normal list
  for (i=0; i<180; i++)
  {
    int ABC[3];                                  // vertex index
    Tl->GetABC(i, ABC);

    int A[3], B[3], C[3];                        // vertex position
    Tl->GetXYZ(ABC[0], A);
    Tl->GetXYZ(ABC[1], B);
    Tl->GetXYZ(ABC[2], C);

    int N[3];                                    // normal to face
    DefUNormal(A, B, C, N);                      // define a normal vector to face (not normalized)
    Normalize(N);                                // normalize vector (to get norm = TRIGO_MUL)
    NList[0] = N[0];                             // store result in list
    NList[1] = N[1];
    NList[2] = N[2];
    NList += 3;
  }

  // -----------------------------------------------------------------
  // init object 1, geosphere with 42 Vertex/80 faces, radius = SPH_R1
  // store intensity into vertex, and uv into faces

  Tl = &Obj3DList[1].Tl;
  if (!Import.InitTL("Mesh/Sph80.msh", Tl, VF_XYZ | VF_I, IF_ABC | IF_UV, SPH_R1, &TextureList, 0))
    return 0;

  // messages
  DisplayMode = 0;                               // used to change display mode (textured/untextured/wireframe)

  // reserve some working memory for internal work of DrawTriangleList() function.
  int MaxVertex;

  if (Obj3DList[0].Tl.VertexListSize > Obj3DList[1].Tl.VertexListSize)
    MaxVertex = Obj3DList[0].Tl.VertexListSize;
  else
    MaxVertex = Obj3DList[1].Tl.VertexListSize;

  if (!VGpu.SetWorkMemory(MaxVertex, 0))         // max triangle list size is 92 vertex, 0 sorted faces (alpha) are used
    return 0;

  return 1;                                      // init success
}

// scan key to change rendering mode
void ScanKeyboard(CVGpu *VGpu)
{
  if (ARMData->Keyboard.KeyPressedId)
  {
    DisplayMode++;
    if (DisplayMode == 4)
      DisplayMode = 0;

    if (DisplayMode == 0)
    {
      VGpu->DrawCfg.TlUnTextured = 1;
      Message = "Untextured Mode";
    }
    else
    if (DisplayMode == 1)
    {
      VGpu->DrawCfg.WireFrame = 1;
      Message = "Wireframe Mode";
    }
    else
    if (DisplayMode == 2)
    {
      VGpu->DrawCfg.NoCulling = 1;
      Message = "Culling OFF";
    }
    else
    if (DisplayMode == 3)
    {
      VGpu->DrawCfg.NoCulling = 0;
      VGpu->DrawCfg.WireFrame = 0;
      VGpu->DrawCfg.TlUnTextured = 0;
      Message = "Textured mode";
    }

    MessageTime = 100;                           // display for 100 frames
    ARMData->Keyboard.KeyPressedId = 0;          // reset
  }
  
  if (MessageTime)
  {
    if (Message)
      VGpu->PrintXY(Message, 0, 30);
    MessageTime--;
  }
}

// main game loop, time is passed, must return 0 to quit
int LoopGame(void)
{
  ScanKeyboard(&VGpu);                           // change display mode ?

  T3DObject *Sph0, *Sph1;
  Sph0 = &Obj3DList[0];                          // fixed sphere
  Sph1 = &Obj3DList[1];                          // satellite sphere

  // ---------------------
  // set sphere 0 position
  Sph0->Matrix.LoadIdentity();                   // init matrix
  Sph0->Matrix.Translate(0, 0, -300);            // define translation on -z axis (camera look on -z direction)
  Sph0->Matrix.RotateX(Sph0->ARotations[0]);     // define xyz ARotations
  Sph0->Matrix.RotateY(Sph0->ARotations[1]);
  Sph0->Matrix.RotateZ(Sph0->ARotations[2]);

  // update Rotations angles
  Sph0->ARotations[0] += 15;
  Sph0->ARotations[1] += 20;
  Sph0->ARotations[2] += 25;

  // ---------------------
  // set sphere 1 position
  Sph1->Matrix.LoadIdentity();
  Sph1->Matrix.Translate(0, 0, -300);            // goto big sphere position
  Sph1->Matrix.RotateX(Sph1->ARotations[0]);     // define rotation around big sphere
  Sph1->Matrix.RotateY(Sph1->ARotations[1]);
  Sph1->Matrix.RotateZ(Sph1->ARotations[2]);
  Sph1->Matrix.Translate(SPH_R0/2 + 60, 0, 0);   // translate around big sphere (radius + 60)
  Sph1->Matrix.RotateX(Sph1->BRotations[0]);     // define self rotation
  Sph1->Matrix.RotateY(Sph1->BRotations[1]);
  Sph1->Matrix.RotateZ(Sph1->BRotations[2]);

  // update Rotations angles
  Sph1->ARotations[0] += -35;
  Sph1->ARotations[1] += 30;
  Sph1->ARotations[2] += -25;
  Sph1->BRotations[0] += 55;
  Sph1->BRotations[1] += -30;
  Sph1->BRotations[2] += 35;

  // ---------------------------------------------------------------
  // define a light power for sphere 1. use frame counter to animate
  int i;
  int Phase = (VGpu.FrameCounter*ANGLE2PI) >> 4;   // 1 complete cycle for 16 frames
  int Power = TRIGO_MUL + SINUS(Phase);            // 0 -> 2*TRIGO_MUL variation
  Power = 80 + ((Power*175) >> (TRIGO_SHIFT + 1)); // define min = 80 + [0..175] range in variation

  // update vertex intensity of satellite sphere
  CTriangleList *Tl = &Sph1->Tl;
  unsigned short *IList = (unsigned short *)&Tl->VertexList[Tl->ViOfs];  // point to first intensity in vertex list
  for (i=0; i<42; i++)
  {
    *IList = Power;
    IList += Tl->VertexStride;
  }

  // ---------------------------------------------------------------
  // define lightening intensity on sphere 0 using sphere 1 position
  int *A = Sph1->Matrix.Pos();                   // get position of sphere 1
  int *B = Sph0->Matrix.Pos();                   // get position of sphere 0

  // light direction
  int AB[3];                                     // light direction
  AB[0] = B[0] - A[0];
  AB[1] = B[1] - A[1];
  AB[2] = B[2] - A[2];
  Normalize(AB);                                 // normalize

  Tl = &Sph0->Tl;
  IList = &Tl->IndexList[Tl->IiOfs];             // get ptr on list intensity of 1st face in triangle list (IF_ABCIUV format used)
  short *NList = Sph0->FNormalList;              // ptr on normal list

#if 0
  // slowest method: rotate each normal vector of object to light
  for (i=0; i<180; i++)
  {
    int N[3];
    N[0] = NList[0];
    N[1] = NList[1];
    N[2] = NList[2];

    Sph0->Matrix.RotateVector(N, N);             // perform only rotation using sphere 0 matrix (ignore translations)
#else
  // fastest method: inverse rotate only light direction, then do not need to rotate each face normal of object to light
  Sph0->Matrix.InvRotateVector(AB, AB);          // perform inverse rotation using sphere 0 transposed matrix
  for (i=0; i<180; i++)
  {
    int N[3];
    N[0] = NList[0];
    N[1] = NList[1];
    N[2] = NList[2];

#endif
    int Dot = AB[0]*N[0] + AB[1]*N[1] + AB[2]*N[2];  // dot product

    int Intensity = 100;                         // define base 'unlighted' intensity
    if (Dot < 0)                                 // triangle is oriented to light
    {
      int di = -Dot >> (TRIGO_SHIFT*2 - 8);      // devide by norm of NList[i] + norm of AB, and scale result by 256, this give ligt effect
      Intensity += (di*Power*155) >> 16;

      // saturate to avoid loop (ex: 257 will define (257 & 0xff) -> intensity = 1)
      if (Intensity > 255)
        Intensity = 255;
    }

    *IList = Intensity;

    IList += Tl->IndexStride;
    NList += 3;
  }

  // ----
  // draw
  VGpu.ViewMatrix = &Sph0->Matrix;                 // change VGpu view matrix (default is &VGpu.Matrix)
  VGpu.DrawTriangleList(&Sph0->Tl, &TextureList);  // note: return value of DrawTriangleList ignored here

  VGpu.ViewMatrix = &Sph1->Matrix;
  VGpu.DrawTriangleList(&Sph1->Tl, &TextureList);

  // text infos
  if ((VGpu.FrameCounter < 300) && !Message)
  {
    VGpu.PrintXY("DrawTriangleList", 0, 15);
    VGpu.PrintXY("+ Face Lightening demo", 0, 30);
    VGpu.PrintXY("Press any key to change display mode", 0, 45);
  }
  VGpu.PrintFps();                               // display framerate
  return !ARMData->Mouse.LButtonDown;            // loop util click/tap on screen occur
}

// blit datas into video memory, and clear virtual screen in dark blue
void BlitScreen(void)
{
  VGpu.FlushFB(COLOR_RGB(0, 0, 64));
}

// free memory
void EndGame(void)
{
  Obj3DList[0].Tl.Delete();
  free(Obj3DList[0].FNormalList);
  Obj3DList[1].Tl.Delete();
  TextureList.Delete();
  VGpu.Delete();
}
#endif // DEMOx
